from .pysaxs3 import pysaxs
from .defineqrois import defineqrois
from .find_center import find_center

class Saxs:

    def __init__(self, ):

        pass


    def pysaxs(*args, **kwargs):
        return pysaxs(*args, **kwargs)

    def defineqrois(*args, **kwargs):
        defineqrois(*args, **kwargs)

    def find_center(*args, **kwargs):
        return find_center(*args, **kwargs)

