from .pyxpcs3 import pyxpcs
from .eventcorrelator3 import eventcorrelator

class Xpcs:

    def __init__():

        pass

    def pyxpcs(*args, **kwargs):
        return pyxpcs(*args, **kwargs)

    def eventcorrelator(*args, **kwargs):
        return eventcorrelator(*args, **kwargs)
